pub struct CurriedFn<D: Domain> { /* private fields */ }Expand description
Wraps a DSL function in a struct that manages currying of the arguments which are fed in one at a time through .apply(). Example: the “+” primitive evaluates to a CurriedFn with arity 2 and empty partial_args. The expression (app + 3) evals to a CurriedFn with vec![3] as the partial_args. The expression (app (app + 3) 4) will evaluate to 7 (since .apply() will fill the last argument, notice that all arguments are filled, and return the result).
Implementations